# $Id: UI_gtk2.pm 2868 2007-07-01 07:17:14Z verthezp $
# $URL: file:///var/lib/svn/intclock/tags/R2_13/lib/Intclock/UI_gtk2.pm $
#___________________________________________________________________
#
# User interface routines for intclock (Gtk2 interface)
#___________________________________________________________________

# Copyright (C) 2004-2007 by Peter Verthez

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

# Written by Peter Verthez, <Peter.Verthez@advalvas.be>.
# Based on Jens-Ulrik Petersen's hsclock

# Note that Intclock::UI_gtk2::conf is defined at the end of this file

package Intclock::UI_gtk2;

use strict;
use diagnostics;
use Glib qw(TRUE FALSE);
use Gtk2;
use Gtk2::Gdk::Keysyms;
use Intclock::TZ;
use Intclock::UI;
use Intclock::Config;
use Encode;
use vars qw($main $aboutBox $popupmenu $active_clock $save_pos_on_exit);

my $true  = 1;
my $false = 0;

my $window_border_width = 3;
my $table_col_spacings  = 6;
my $table_row_spacings  = 3;
my $label_x_padding     = 3;
my $label_y_padding     = 3;
my $codeset;

sub _get_locale_codeset {
  eval {
    require I18N::Langinfo;
    I18N::Langinfo->import(qw(langinfo CODESET));
    $codeset = langinfo(CODESET());
  };
  if ($@) {
    print STDERR "WARNING: Could not find encoding for current locale\n";
  }
}

sub start {
  my ($optval, $progname, $about, $language, $geometry, $save) = @_;
  $save_pos_on_exit = $save;
  init Gtk2;
  _get_locale_codeset();
  _create_ui($optval, $progname, $about, $language, $geometry);
  main Gtk2;
}

sub _create_ui {
  my ($optval, $progname, $about, $language, $geometry) = @_;
  
  # main window
  $main = new Gtk2::Window("toplevel");
  $main->set_title($progname);
  $main->set_border_width($window_border_width);
  $main->set("allow-shrink", $false);
  $main->set("allow-grow", $true);
  $main->set_resizable($false);
  my $datadir = Intclock::Config::get_datadir();
  $main->set_default_icon_from_file("$datadir/intclock.png");
  $main->signal_connect("delete_event", \&_cb_exit_ui_as_ok, [$optval]);

  _create_menu($optval, $progname, $about, $language);

  $main->add_events("button_press_mask");
  $main->signal_connect("button_press_event", \&_cb_button_press);
  $main->signal_connect("popup_menu", \&_cb_popup_menu);

  # table for the clocks
  _create_clock_table($optval);

  # show all widgets
  $main->show_all();
  
  # position window
  if ($geometry) {
    my $orig_geometry = $geometry;
    # ignore width, height part of geometry
    $geometry =~ s/^[0-9]+x[0-9]+//;
    my $success = $main->parse_geometry($geometry);
    print STDERR "Unable to parse geometry string '$orig_geometry'\n"
	if not $success;
  }  
}

my @clock_menuitems = ();

sub _create_clock_menu_items {
  my ($config) = @_;
  my $t = time();
  my $menu_item;
  my $menu_group;
  $active_clock = 0;
  {
    $menu_item = new Gtk2::MenuItem();
    $popupmenu->prepend($menu_item);
    $menu_item->show();
    push @clock_menuitems, [$menu_item, undef];
  }
  my $clock_nr = scalar @{$config->{"zones"}};
  foreach my $zone (reverse @{$config->{"zones"}}) {
    my ($zonename, $username, $default) = @$zone;
    my $tz_a = Intclock::TZ::in_timezone($t, $zonename);
    my $zonetext = Intclock::UI::timezone_label($zone, $tz_a);
    
    $menu_item = Gtk2::RadioMenuItem->new($menu_group, $zonetext);
    $menu_item->signal_connect("activate", \&_cb_activate_clock,
			       [--$clock_nr, $config]);
    $menu_group = $menu_item if not defined $menu_group;
    $popupmenu->prepend($menu_item);
    $menu_item->show();
    push @clock_menuitems, [$menu_item, $zone];
    $menu_item->set_active($true) if $default;
  }
  my ($first_item) = @{$clock_menuitems[-1]};
  $first_item->set_active($true);
}

sub _cb_activate_clock {
  my ($menuitem, $data) = @_;
  my ($clocknr, $config) = @$data;
  if ($menuitem->get_active() and $active_clock != $clocknr) {
    $active_clock = $clocknr;
    _cb_redraw_clocks($config, "foractivate");
  }
}

sub _remove_clock_menu_items {
  foreach my $item (@clock_menuitems) {
    my ($menuitem, $zone) = @$item;
    $popupmenu->remove($menuitem);
  }
  @clock_menuitems = ();
}

sub _redraw_clock_menu_items {
  my ($config) = @_;
  _remove_clock_menu_items();
  _create_clock_menu_items($config);
}

sub _create_clock_table {
  my ($optval, $foractivate) = @_;
  # determine nr of rows and columns for table
  my ($rows, $columns);
  my $nrZones = scalar(@{$optval->{"zones"}});
  my ($layout, $layout_arg) = @{$optval->{"layout"}};
  if ($layout eq "horizontal") {
    $rows = 1; $columns = $nrZones;
    _remove_clock_menu_items() unless $foractivate;
  }
  elsif ($layout eq "columns") {
    use integer;
    $columns = $layout_arg;
    $columns = $nrZones if $columns > $nrZones;
    $rows    = ($nrZones - 1) / $columns + 1;
    _remove_clock_menu_items() unless $foractivate;
  }
  elsif ($layout eq "vertical") {
    $rows = $nrZones; $columns = 1;
    _remove_clock_menu_items() unless $foractivate;
  }
  elsif ($layout eq "oneclock") {
    $rows = 1; $columns = 1;
    _redraw_clock_menu_items($optval) unless $foractivate;
  }

  my $table = new Gtk2::Table($rows, $columns, $true);
  $table->set_col_spacings($table_col_spacings);
  $table->set_row_spacings($table_row_spacings);
  $main->add($table);

  _setup_clocks($table, $layout, $columns, $optval->{"zones"});
}

my $decode_warning_shown = 0;

sub _fix_timestr {
  my ($str) = @_;
  if (defined $codeset) {
    eval {
      $str = Encode::decode($codeset, $str);
    };
    if ($@) {
      print STDERR "WARNING: Decoding from current encoding failed\n"
	  if not $decode_warning_shown;
      $decode_warning_shown = 1;
    }
  }
  return $str;
}

sub _setup_clock {
  my ($table, $columns, $zone, $t, $i) = @_;
  my ($zonename) = @$zone;
  my $tz_a = Intclock::TZ::in_timezone($t, $zonename);
  my $tz_t = Intclock::UI::date_time_string($tz_a);
  $tz_t = _fix_timestr($tz_t);
  my $zonetext = Intclock::UI::timezone_label($zone, $tz_a);
  
  my $frame = new Gtk2::Frame($zonetext);
  $table->attach_defaults($frame,
			  $i % $columns, $i % $columns + 1,
			  $i / $columns, $i / $columns + 1);
  
  my $label = new Gtk2::Label($tz_t);
  $label->set_justify("center");
  $label->set_padding($label_x_padding, $label_y_padding);
  $frame->add($label);
  return [$label, $frame, $zone];
}

sub _setup_clocks {
  my ($table, $layout, $columns, $zones) = @_;
  my $t = time();
  Intclock::UI::debug($t);
  my @clocks;
  if ($layout eq "oneclock") {
    my $zone = $zones->[$active_clock];
    push @clocks, _setup_clock($table, $columns, $zone, $t, 0);
  }
  else {
    my $i = 0;
    foreach my $zone (@$zones) {
      push @clocks, _setup_clock($table, $columns, $zone, $t, $i);
      $i++;
    }
  }
  _set_timer(\@clocks, $layout);
}

sub _cb_redraw_clocks {
  my ($config, $foractivate) = @_;
  my ($table) = $main->get_children();
  $main->remove($table);

  _create_clock_table($config, $foractivate);
  $main->show_all();
}

sub _cb_window_pos {
  my ($button) = @_;
  $save_pos_on_exit = ($button->get_active() ? 1 : 0);
}

sub _set_timer {
  my ($clocks) = @_;
  my $timer = Intclock::UI::get_timer_interval();
  Glib::Timeout->add($timer, \&_cb_update_clocks, $clocks);
}

sub _cb_update_clocks {
  my ($clocks) = @_;
  my $t = time();
  Intclock::UI::debug($t);
  foreach my $clock (@$clocks) {
    my ($label, $frame, $zone) = @$clock;
    my ($zonename) = @$zone;
    my $tz_a = Intclock::TZ::in_timezone($t, $zonename);
    my $tz_t = Intclock::UI::date_time_string($tz_a);
    my $zonetext = Intclock::UI::timezone_label($zone, $tz_a);
    $tz_t = _fix_timestr($tz_t);

    $label->set_text($tz_t);
    $frame->set_label($zonetext);
  }
  _set_timer($clocks);
  return $false;
}

sub _cb_exit_ui_as_ok {
  my ($widget, @args) = @_;
  # ignore event if from destroy
  shift @args if (ref $widget eq "Gtk2::Window");
  my $data = shift @args;
  my ($config) = @$data;
  if ($save_pos_on_exit) {
    $config->{"position"} = [ $main->get_position() ];
  }
  else {
    delete $config->{"position"};
  }
  Intclock::Config::save($config);
  Gtk2->main_quit();
  return $false;
}

sub _create_menu {
  my ($config, $progname, $about, $language) = @_;
  $popupmenu = new Gtk2::Menu();
  my $accelgroup = new Gtk2::AccelGroup();
  $main->add_accel_group($accelgroup);
  $popupmenu->set_accel_group($accelgroup);
  my $menu_item;
  {
    $menu_item = Gtk2::ImageMenuItem->new_from_stock("gtk-preferences");
    $popupmenu->append($menu_item);
    $menu_item->signal_connect("activate",
			       \&Intclock::UI_gtk2::conf::show,
			       [$main, $config, $progname, $language,
				\&_cb_redraw_clocks, \&_cb_window_pos]);
    $menu_item->show();
  }
  {
    $menu_item = new Gtk2::MenuItem();
    $popupmenu->append($menu_item);
    $menu_item->show();
  }
  {
    $menu_item = new Gtk2::MenuItem($language->maketext("_About"));
    $popupmenu->append($menu_item);
    $menu_item->signal_connect("activate", \&_cb_about,
			       [$language->maketext("About [_1]", $progname),
				$about]);
    $menu_item->show();
  }
  {
    $menu_item = Gtk2::ImageMenuItem->new_from_stock("gtk-quit");
    $menu_item->set_accel_path("<$progname>/Quit");
    my $key = $Gtk2::Gdk::Keysyms{Q};
    my $mask = qw(control-mask);
    Gtk2::AccelMap->add_entry("<$progname>/Quit", $key, $mask);
    $popupmenu->append($menu_item);
    $menu_item->signal_connect("activate", \&_cb_exit_ui_as_ok, [$config]);
    $menu_item->show();
  }
}

sub _my_menu_positioning_func {
  my ($menu, $x, $y, $main) = @_;
  my ($win_x, $win_y) = $main->get_position();
  my ($width, $height) = $main->get_size();
  if ($x >= $win_x and $x <= $win_x + $width
      and $y >= $win_y and $y <= $win_y + $height) {
    return ($x, $y, 1);
  }
  else {
    return ($win_x + $width, $win_y + $height, 1);
  }
}

sub _cb_button_press {
  my ($button, $event) = @_;
  if (($event->button == 3)) {
    $popupmenu->popup(undef, undef, undef, undef, $event->button, 0);
    return $true;
  }
  
  return $false;
}

sub _cb_popup_menu {
  my ($main) = @_;
  $popupmenu->popup(undef, undef, \&_my_menu_positioning_func, $main, 0, 0);
  return $true;
}

sub _cb_about {
  my ($button, $data) = @_;
  if (not defined $aboutBox) {
    my ($title, $contents) = @$data;
    $aboutBox = Gtk2::MessageDialog->new (undef, [], 'info', 'close',
					  $contents);
    $aboutBox->set_title($title);
    $aboutBox->signal_connect (response => sub {$_[0]->hide});
  }
  $aboutBox->show;
}

1;

###############################################################################
#  configuration window
###############################################################################

package Intclock::UI_gtk2::conf;

use strict;
use diagnostics;
use Gtk2;
use POSIX qw(strftime);
use Intclock::Config;
use Intclock::TZ;
use Intclock::UI;
use vars qw($dialog $tzTree $tzList $tzCommentsEntry $tzNameEntry
	    $tzNameLabel $tzNameButton
	    $addTzButton $delTzButton $upTzButton $downTzButton $nameButton
	    $userNameButton $secondsButton $hour24Button $dateButton
	    $customFormatButton
	    $tzButton $vowelButton $horizButton $vertButton $oneClockButton
	    $colButton $colSpin $timeFormatCombo $timeFormatLabel
	    $selectedTzInTree $selectedTzInList @specialTimezones
	    $savePosButton
	    $selectedL $numInList $initialized);

use constant TREE_COLUMN_TIMEZONE => 0;
use constant TREE_COLUMN_REAL_TIMEZONE => 1;

use constant COLUMN_TIMEZONE  => 0;
use constant COLUMN_USERNAME  => 1;
use constant COLUMN_EDITABLE  => 2;
use constant COLUMN_REAL_TIMEZONE  => 3;

BEGIN {
  @specialTimezones = ("UTC");
  $selectedL = 0;
  $numInList = 0;
  $initialized = 0;
}

my $spacing = 4;
my $controlspacing = 12;
my $defopts = ['fill', 'expand'];
my $noexpand = ['fill'];

sub _make_title_label {
  my ($language, $text, $container, $padding) = @_;
  my $label = new Gtk2::Label();
  $label->set_markup("<b>" . $language->maketext($text) . "</b>");
  $label->set_justify('left');
  $label->set_alignment(0.0, 0.5);
  $label->set_padding($controlspacing, 0) if defined $padding;
  if (ref $container eq "ARRAY") {
    my ($table, @coords) = @$container;
    $table->attach($label, @coords, $defopts, $noexpand, $controlspacing, 0);
  }
  else {
    my $box = $container;
    $box->pack_start($label, $false, $true,
		     (ref $container eq "Gtk2::HBox") ? $controlspacing
		                                      : $spacing);
  }
  $label->show();
  return $label;
}

sub _make_vbuttonbox {
  my ($table, @coords) = @_;
  my $vbuttonbox = new Gtk2::VButtonBox();
  $table->attach($vbuttonbox, @coords, $noexpand, $defopts,
		  $controlspacing, $spacing);
  $vbuttonbox->set_layout('start');
  $vbuttonbox->set_spacing($spacing * 2);
  $vbuttonbox->show();
  return $vbuttonbox;
}

sub _make_button {
  my ($stock, $vbox, $callback, $args) = @_;
  my $button = Gtk2::Button->new_from_stock($stock);
  $button->set_sensitive($false);
  $button->signal_connect("clicked", $callback, $args);
  $vbox->pack_start($button, $true, $true, $spacing);
  $button->show();
  return $button;
}

sub _make_radiobutton {
  my ($group, $language, $text, $container, $callback, $args) = @_;
  my ($table, @coords) = @$container;
  my $radiobutton = new Gtk2::RadioButton($group, $language->maketext($text));
  $radiobutton->signal_connect("toggled", $callback, $args);
  $table->attach($radiobutton, @coords,
		 $defopts, $defopts, $controlspacing * 2, 4);
  $radiobutton->show();
  return $radiobutton;
}

sub _make_checkbutton {
  my ($language, $text, $container, $callback, $args) = @_;
  my ($table, @coords) = @$container;
  my $checkbutton = new Gtk2::CheckButton($language->maketext($text));
  $checkbutton->signal_connect("toggled", $callback, $args);
  $table->attach($checkbutton, @coords,
		 $defopts, $defopts, $controlspacing * 2, 4);
  $checkbutton->show();
  return $checkbutton;
}

sub _init {
  my ($main, $config, $progname, $language, $redrawfunc, $windowposfunc) = @_;

  ###
  ### General stuff: dialog, notebook
  ### 
    
  $dialog = new Gtk2::Dialog($language->maketext("Preferences for [_1]",
						 $progname),
			     $main, [ ]);
  $dialog->set_position("center");
  # Block deletion of the window:
  $dialog->signal_connect("delete_event", sub { return $true });

  my $notebook = new Gtk2::Notebook();
  $dialog->vbox->pack_start($notebook, $true, $true, $spacing);
  $notebook->show();

  ###
  ### First tab
  ### 
    
  my $vbox_page1 = new Gtk2::VBox($false, $spacing);
  $vbox_page1->show();
  $notebook->append_page($vbox_page1, $language->maketext("Timezones"));

  # vbox_page1.table1: zone table
  my $table1 = new Gtk2::Table(6, 2, $false);
  $table1->set_row_spacings(10);
  $table1->set_row_spacing(0, 0);
  $table1->set_col_spacings(10);
  $vbox_page1->pack_start($table1, $true, $true, $controlspacing);
  $table1->show();

  # vbox_page1.table1.label11
  _make_title_label($language, "Available Timezones", [$table1, 0, 1, 0, 1]);

  # vbox_page1.table1.scrolled11: scrolled window for timezone tree
  my $scrolled11 = new Gtk2::ScrolledWindow(undef, undef);
  $scrolled11->set_size_request(400, 150);
  $table1->attach($scrolled11, 0, 1, 1, 2, $defopts, $defopts,
		  $controlspacing * 2, $spacing);
  $scrolled11->set_policy('automatic', 'automatic');
  $scrolled11->show();

  # vbox_page1.table1.scrolled11.tzTree: timezone tree
  my $tzTreeModel = new Gtk2::TreeStore("Glib::String",
					"Glib::String");
  $tzTree = new Gtk2::TreeView($tzTreeModel);
  $tzTree->set_enable_search($false);
  $tzTree->set_headers_visible($false);
  $tzTree->get_selection->set_mode('GTK_SELECTION_SINGLE');
  $tzTree->get_selection->signal_connect('changed', \&_cb_select_zone_in_tree);
  $scrolled11->add_with_viewport($tzTree);

  my $renderer = new Gtk2::CellRendererText();
  $renderer->set(xalign => 0.0);
  my $col_offset = $tzTree->insert_column_with_attributes 
      (-1, "Timezone", $renderer, text => TREE_COLUMN_TIMEZONE);
  my $column = $tzTree->get_column($col_offset - 1);
  $column->set_clickable($true);
    
  $tzTree->show();

  # vbox_page1.table1.vbuttonbox1: buttons for timezone tree
  my $vbuttonbox11 = _make_vbuttonbox($table1, 1, 2, 1, 2);

  # add button
  $addTzButton = _make_button("gtk-add", $vbuttonbox11, \&_cb_add_timezone,
			      [ $main, $config, $redrawfunc ]);

  # vbox_page1.table1.hbox_info: hbox for timezone info
  my $hbox_info = new Gtk2::HBox($false, $spacing);
  $table1->attach($hbox_info, 0, 2, 2, 3, $defopts, $noexpand, 0, 0);
  $hbox_info->show();

  # vbox_page1.table1.hbox_info.label13
  _make_title_label($language, "Info for Selected Zone:", $hbox_info);

  # vbox_page1.table1.hbox_info.tzCommentsEntry
  $tzCommentsEntry = new Gtk2::Entry();
  $hbox_info->pack_start($tzCommentsEntry, $true, $true, $spacing);
  $tzCommentsEntry->set_editable($false);
  $tzCommentsEntry->show();

  # vbox_page1.table1.label14
  _make_title_label($language, "Selected Timezones", [$table1, 0, 1, 4, 5]);

  # scrolled window for timezone selection
  my $scrolled14 = new Gtk2::ScrolledWindow(undef, undef);
  $scrolled14->set_size_request(400, 150);
  $table1->attach($scrolled14, 0, 1, 5, 6, $defopts, $defopts,
		  $controlspacing * 2, 0);
  $scrolled14->set_policy('automatic', 'automatic');
  $scrolled14->show();

  # timezone selection
  my $tzListModel = new Gtk2::ListStore("Glib::String",
					"Glib::String",
					"Glib::Boolean",
					"Glib::String");
  $tzList = new Gtk2::TreeView($tzListModel);
  $tzList->set_enable_search($false);
  $tzList->get_selection->set_mode('GTK_SELECTION_SINGLE');
  $tzList->get_selection->signal_connect('changed', \&_cb_select_zone_in_list);
  $scrolled14->add_with_viewport($tzList);

  $renderer = new Gtk2::CellRendererText();
  $renderer->set(xalign => 0.0);
  $col_offset = $tzList->insert_column_with_attributes 
      (-1, $language->maketext("Timezone"), $renderer,
       text => COLUMN_TIMEZONE);
  $column = $tzList->get_column($col_offset - 1);
  $column->set_clickable($true);
  $column->set_resizable($true);
  
  $renderer = new Gtk2::CellRendererText();
  $renderer->signal_connect ("edited", \&_cb_change_timezone_user_name,
			     [ $tzListModel, $main, $config, $redrawfunc ]);
  $renderer->set(xalign => 0.0);
  $col_offset = $tzList->insert_column_with_attributes 
      (-1, $language->maketext("User Defined Name"), $renderer,
       text => COLUMN_USERNAME,
       editable => COLUMN_EDITABLE);
  $column = $tzList->get_column($col_offset - 1);
  $column->set_clickable($true);
  $column->set_resizable($true);
  
  $tzList->show();
    
  # vbox_page1.table1.vbuttonbox2: buttons for timezone list
  my $vbuttonbox12 = _make_vbuttonbox($table1, 1, 2, 5, 6);

  # remove, up, down button
  $delTzButton = _make_button("gtk-remove", $vbuttonbox12,
			      \&_cb_remove_timezone,
			      [ $main, $config, $redrawfunc ]);
  
  $upTzButton = _make_button("gtk-go-up", $vbuttonbox12,
			     \&_cb_move_timezone_up,
			     [ $main, $config, $redrawfunc ]);
  
  $downTzButton = _make_button("gtk-go-down", $vbuttonbox12,
			       \&_cb_move_timezone_down,
			       [ $main, $config, $redrawfunc ]);
  
  ###
  ### Second tab
  ### 
    
  my $vbox_page2 = new Gtk2::VBox($false, $spacing);
  $vbox_page2->show();
  $notebook->append_page($vbox_page2, $language->maketext("Options"));

  # vbox_page2.label21 (padding because not in table)
  _make_title_label($language, "Clock Label", $vbox_page2, "pad");

  # buttons
  my $table21 = new Gtk2::Table(4, 2, $false);
  $table21->set_col_spacing(0, 30);
  $vbox_page2->pack_start($table21, $false, $true, $spacing);

  $tzButton
      = _make_radiobutton(undef, $language, "Use _timezone abbreviation",
			  [$table21, 0, 2, 0, 1], 
			  \&_cb_toggle_display_option,
			  [ $main, $config, $redrawfunc ]);

  $nameButton
      = _make_radiobutton($tzButton, $language, "Use c_ity/country name",
			  [$table21, 0, 2, 1, 2], 
			  \&_cb_toggle_display_option,
			  [ $main, $config, $redrawfunc ]);

  $vowelButton
      = _make_checkbutton($language, "Drop _vowels",
			  [$table21, 1, 2, 2, 3],
			  \&_cb_apply_change, [ $main, $config, $redrawfunc ]);
  $vowelButton->set_sensitive($false);

  $userNameButton
      = _make_radiobutton($tzButton, $language, "Use _user defined name",
			  [$table21, 0, 2, 3, 4],
			  \&_cb_toggle_display_option,
			  [ $main, $config, $redrawfunc ]);
  $table21->show();
  
  # vbox_page1.label (padding because not in table)
  _make_title_label($language, "Time Format", $vbox_page2, "pad");

  # buttons for time format
  my $table22 = new Gtk2::Table(4, 2, $false);
  $table22->set_col_spacing(0, 30);
  $vbox_page2->pack_start($table22, $false, $true, $spacing);
  $table22->show();

  $secondsButton
     = _make_checkbutton($language, "Show _seconds",
			 [$table22, 0, 2, 0, 1],
			 \&_cb_set_time_format, [$main, $config, $redrawfunc]);
  
  $hour24Button
     = _make_checkbutton($language, "Show 24 _hour format",
			 [$table22, 0, 2, 1, 2],
			 \&_cb_set_time_format, [$main, $config, $redrawfunc]);

  $dateButton
     = _make_checkbutton($language, "Show _date",
			 [$table22, 0, 2, 2, 3],
			 \&_cb_set_time_format, [$main, $config, $redrawfunc]);

  # button for custom format
  $customFormatButton
     = _make_checkbutton($language, "Use custom _format:",
			 [$table22, 0, 1, 3, 4],
			 \&_cb_set_time_format,
			 [$main, $config, $redrawfunc, "custom"]);

  # combobox for selecting time format
  $timeFormatCombo = new Gtk2::Combo();
  $timeFormatCombo->set_popdown_strings(@Intclock::UI::timeformats);
  $timeFormatCombo->entry->signal_connect("changed",
					  \&_cb_apply_change,
					  [ $main, $config, $redrawfunc ]);
  $table22->attach($timeFormatCombo, 1, 2, 3, 4,
		   $defopts, $defopts, $controlspacing * 2, 4);
  $timeFormatCombo->show();

  # vbox_page1.label22 (padding because not in table)
  _make_title_label($language, "Clock Table Layout", $vbox_page2, "pad");

  # buttons
  my $table23 = new Gtk2::Table(4, 2, $false);
  $vbox_page2->pack_start($table23, $false, $true, $spacing);
  $table23->show();
  
  $oneClockButton
      = _make_radiobutton(undef, $language, "_One clock at a time",
			  [$table23, 0, 2, 0, 1], 
			  \&_cb_apply_change, [ $main, $config, $redrawfunc ]);

  $vertButton
      = _make_radiobutton($oneClockButton, $language,
			  "_Vertical layout (1 column)",
			  [$table23, 0, 2, 1, 2], 
			  \&_cb_apply_change, [ $main, $config, $redrawfunc ]);

  $horizButton
      = _make_radiobutton($oneClockButton, $language,
			  "_Horizontal layout (1 row)",
			  [$table23, 0, 2, 2, 3], 
			  \&_cb_apply_change, [ $main, $config, $redrawfunc ]);

  $colButton
      = _make_radiobutton($oneClockButton, $language,
			  "_Number of columns:",
			  [$table23, 0, 1, 3, 4], 
			  \&_cb_toggle_column_entry,
			  [ $main, $config, $redrawfunc ]);
  
  my $adj = new Gtk2::Adjustment(2, 2, 10, 1, 1, 1);
  $colSpin = new Gtk2::SpinButton($adj, 0, 0);
  $colSpin->signal_connect("value_changed", \&_cb_apply_change,
			   [ $main, $config, $redrawfunc ]);
  $table23->attach($colSpin, 1, 2, 3, 4,
		   $defopts, $defopts, $controlspacing * 2, 4);
  $colSpin->show();
  $colSpin->set_sensitive($false);
  
  ###
  ### Third tab
  ### 
    
  my $vbox_page3 = new Gtk2::VBox($false, $spacing);
  $vbox_page3->show();
  $notebook->append_page($vbox_page3, $language->maketext("Window"));

  # vbox_page3.label (padding because not in table)
  _make_title_label($language, "Window Position", $vbox_page3, "pad");

  # buttons
  my $table31 = new Gtk2::Table(1, 1, $false);
  $vbox_page3->pack_start($table31, $false, $true, $spacing);
  $table31->show();
  
  $savePosButton
     = _make_checkbutton($language, "_Save window position on exit",
			 [$table31, 0, 1, 0, 1], $windowposfunc);

  ###
  ### Button area
  ### 
    
  # Close button
  my $closeButton = Gtk2::Button->new_from_stock("gtk-close");
  $dialog->action_area->pack_start($closeButton, $true, $true, 0);
  $closeButton->signal_connect("clicked", sub { $dialog->hide() });
  $closeButton->show();
  
  _fill_timezone_tree();
}

sub show {
  my ($button, $data) = @_;
  my ($main, $config, $progname, $language, $redrawfunc, $windowposfunc)
      = @$data;
  $initialized = 0;
  _init($main, $config, $progname, $language, $redrawfunc, $windowposfunc)
      if not defined $dialog;
  _ui_load_config($config);
  $dialog->show();
  $initialized = 1;
}

sub _apply {
  my ($main, $config, $redrawfunc) = @_;
  if ($initialized and _ui_save_config($main, $config)) {
    Intclock::UI::set_config($config);
    &{$redrawfunc}($config);
  }
}

sub _add_timezone_tree_branch {
  my ($treemodel, $parentiter, $label) = @_;
  
  my $iter = $treemodel->append($parentiter);
  $treemodel->set($iter, TREE_COLUMN_TIMEZONE, $label);
  return $iter;
}

sub _add_timezone_tree_leaf {
  my ($treemodel, $parentiter, $label, $real_zone) = @_;

  my $iter = $treemodel->append($parentiter);
  $treemodel->set($iter, TREE_COLUMN_TIMEZONE, $label);
  $treemodel->set($iter, TREE_COLUMN_REAL_TIMEZONE, $real_zone);
}

sub _tz_text {
  my (@tzAbbr) = @_;
  if ($tzAbbr[0] eq $tzAbbr[1]) {
    return $tzAbbr[0];
  }
  else {
    return "$tzAbbr[0]/$tzAbbr[1]";
  }
}

sub _zones_by_longitude {
  my ($zone_l) = @_;
  my @keys = keys %$zone_l;
  @keys = sort { $zone_l->{$a}->{"long"} <=> $zone_l->{$b}->{"long"} } @keys;
  return @keys;
}

sub _fill_timezone_tree {
  my ($treeitem, $subtree, $subsubtree);
  use locale;
  use POSIX qw(locale_h);
  setlocale(LC_COLLATE, "");
  my $model = $tzTree->get_model();
  my ($zonehash) = Intclock::TZ::zone_info();

  for my $timezone (@specialTimezones) {
    _add_timezone_tree_leaf($model, undef, $timezone, $timezone);
  }

  foreach my $cont (sort keys %$zonehash) {
    my $subtree = _add_timezone_tree_branch($model, undef, $cont);
    my $zone_c = $zonehash->{$cont};
    foreach my $country (sort keys %$zone_c) {
      my $zone_l = $zone_c->{$country};
      if (scalar keys %$zone_l == 1) {
	my ($zone) = keys %$zone_l;
	my $zoneinfo = $zonehash->{$cont}->{$country}->{$zone};
	my $continent = $zoneinfo->{"continent"};
	my $tzAbbr = _tz_text(Intclock::TZ::get_timezone_names("$continent/$zone"));
	_add_timezone_tree_leaf($model, $subtree,
				"$country [$tzAbbr]",
				"$cont|$country|$zone");
      }
      else {
	$subsubtree = _add_timezone_tree_branch($model, $subtree, $country);
	foreach my $zone (_zones_by_longitude $zone_l) {
	  my $zoneinfo = $zonehash->{$cont}->{$country}->{$zone};
	  my $continent = $zoneinfo->{"continent"};
	  my $tzAbbr = _tz_text(Intclock::TZ::get_timezone_names("$continent/$zone"));
	  my $tzone = $zone;
	  if (exists $zoneinfo->{"translated_zone"}) {
	    $tzone = $zoneinfo->{"translated_zone"};
	  }
	  $tzone =~ s|^.*?/||;
	  $tzone =~ s/_/ /g;
	  _add_timezone_tree_leaf($model, $subsubtree, "$tzone [$tzAbbr]",
				  "$cont|$country|$zone");
	}
      }
    }
  }
}

sub _cb_toggle_display_option {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  $vowelButton->set_sensitive($nameButton->get_active());
  $vowelButton->set_active($false);
  _apply($main, $config, $redrawfunc);
}

sub _cb_toggle_column_entry {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  $colSpin->set_sensitive($widget->get_active());
  _apply($main, $config, $redrawfunc);
}

sub _cb_apply_change {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  _apply($main, $config, $redrawfunc);
}

sub _cb_select_zone_in_tree {
  my ($treeselection) = @_;
  my ($model, $iter) = $treeselection->get_selected();
  if (not defined $model or defined $model->iter_children($iter)) {
    undef $selectedTzInTree;
    $tzCommentsEntry->set_text("");
    $addTzButton->set_sensitive($false);
  }
  else {
    my @path = ();
    my $entry = $model->get($iter, TREE_COLUMN_REAL_TIMEZONE);
    my ($cont, $country, $zone);
    if ($entry !~ /\|/) {
      $zone = $entry;
    }
    else  {
      ($cont, $country, $zone) = split /\|/, $entry;
    }
    
    if (defined $cont) {
      my ($zonehash) = Intclock::TZ::zone_info();
      my $zoneinfo = $zonehash->{$cont}->{$country}->{$zone};
      my $continent = $zoneinfo->{"continent"};
      $selectedTzInTree = "$continent/$zone";
      my $comment = $zoneinfo->{"comments"};
      $comment ||= $country;
      $tzCommentsEntry->set_text($comment);
    }
    else {
      $selectedTzInTree = $zone;
    }
    $addTzButton->set_sensitive($true);
  }
  if (defined $model) {
    my $treeselection = $tzList->get_selection();
    $treeselection->unselect_all();
  }
}

sub _cb_select_zone_in_list {
  my ($listselection) = @_;
  my ($model, $iter) = $listselection->get_selected();
  if (defined $model) {
    my ($row) = $model->get_path($iter)->get_indices();
    $selectedTzInList = $row;
    $delTzButton->set_sensitive($numInList > 1 ? $true : $false);
    $upTzButton->set_sensitive($row > 0 ? $true : $false);
    $downTzButton->set_sensitive($row < $numInList - 1 ? $true : $false);
    my $treeselection = $tzTree->get_selection();
    $treeselection->unselect_all();
  }
  else {
    undef $selectedTzInList;
    $delTzButton->set_sensitive($false);
    $upTzButton->set_sensitive($false);
    $downTzButton->set_sensitive($false);
  }
}

sub _add_timezone_helper {
  my ($tz) = @_;
  $numInList++;
  my $model = $tzList->get_model();
  my $iter = $model->append();
  my ($zoneinfo, $invzoneinfo) = Intclock::TZ::zone_info();
  my $tzname = $tz->[0];
  if (exists $invzoneinfo->{$tz->[0]}) {
    my ($country, $city, $countryhash) = @{$invzoneinfo->{$tz->[0]}};
    if (scalar(keys %$countryhash) > 1) {
      $tzname = "$country - $city";
    }
    else {
      $tzname = $country;
    }
  }
  $model->set($iter,
	      COLUMN_TIMEZONE, $tzname,
	      COLUMN_USERNAME, $tz->[1],
	      COLUMN_EDITABLE, $true,
	      COLUMN_REAL_TIMEZONE, $tz->[0]);
}

sub _cb_add_timezone {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  _add_timezone_helper([$selectedTzInTree, ""]);
  my $treeselection = $tzTree->get_selection();
  $treeselection->unselect_all();
  _apply($main, $config, $redrawfunc);
}

sub _cb_remove_timezone {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  my $treeselection = $tzList->get_selection();
  $numInList--;
  my $model = $tzList->get_model();
  $model->remove($model->iter_nth_child(undef, $row));
  _apply($main, $config, $redrawfunc);
  if ($row < $numInList) {
    $treeselection->select_iter($model->iter_nth_child(undef, $row));
  }
}

sub _cb_move_timezone_up {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  my $treeselection = $tzList->get_selection();
  my ($model, $iter) = $treeselection->get_selected();
  $treeselection->unselect_all();
  $model->move_before($iter, $model->iter_nth_child(undef, $row - 1));
  $treeselection->select_iter($iter);
  _apply($main, $config, $redrawfunc);
}

sub _cb_move_timezone_down {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc) = @$data;
  my $row = $selectedTzInList;
  my $treeselection = $tzList->get_selection();
  my ($model, $iter) = $treeselection->get_selected();
  $treeselection->unselect_all();
  $model->move_after($iter, $model->iter_next($iter));
  $treeselection->select_iter($iter);
  _apply($main, $config, $redrawfunc);
}

sub _cb_change_timezone_user_name {
  my ($cell, $path_string, $new_text, $data) = @_;
  my ($model, $main, $config, $redrawfunc) = @$data;

  my $path = Gtk2::TreePath->new_from_string($path_string);
  my $iter = $model->get_iter($path);
  $model->set($iter, COLUMN_USERNAME, $new_text);
  _apply($main, $config, $redrawfunc);
}

sub _calculate_time_format_buttons {
  my ($initial) = @_;
  my $format = $timeFormatCombo->entry->get_text();
  my $predefined = exists $Intclock::UI::tf_configs{$format};

  if ($initial) {
    $customFormatButton->set_active($predefined ? $false : $true);
    $timeFormatCombo->set_sensitive($predefined ? $false : $true);
  }
  
  if ($predefined) {
    my $config = $Intclock::UI::tf_configs{$format};
    $secondsButton->set_active($config->[0] ? $true : $false);
    $hour24Button->set_active($config->[1] ? $true : $false);
    $dateButton->set_active($config->[2] ? $true : $false);
  }
  else {
    $secondsButton->set_sensitive($false);
    $hour24Button->set_sensitive($false);
    $dateButton->set_sensitive($false);
  }
}

sub _recalculate_time_format {
  my $timeformat = "";
  my $seconds = "";
  if ($dateButton->get_active()) {
    $timeformat .= "%a %d%n";
  }
  if ($secondsButton->get_active()) {
    $seconds = ":%S";
  }
  if ($hour24Button->get_active()) {
    $timeformat .= "%H:%M$seconds";
  }
  else {
    $timeformat .= "%I:%M$seconds %p";
  }
  $timeFormatCombo->entry->set_text($timeformat);
}

sub _cb_set_time_format {
  my ($widget, $data) = @_;
  my ($main, $config, $redrawfunc, $custom) = @$data;
  if ($custom and $custom eq "custom") {
    my $state = $customFormatButton->get_active();
    $timeFormatCombo->set_sensitive($state);
    $secondsButton->set_sensitive($state ? $false : $true);
    $hour24Button->set_sensitive($state ? $false : $true);
    $dateButton->set_sensitive($state ? $false : $true);
    if (not $state) {
      _calculate_time_format_buttons($false);
    }
  }
  else {
    _recalculate_time_format();
  }
  _apply($main, $config, $redrawfunc);
}

sub _ui_load_config {
  my ($config) = @_;
  $tzList->get_model()->clear();
  if ($config->{"zones"}) {
    foreach my $timezone (@{$config->{"zones"}}) {
      _add_timezone_helper($timezone);
    }
  }
  if ($config->{"name"}) {
    if ($config->{"name"} == 1) {
      $nameButton->set_active($true);
    }
    elsif ($config->{"name"} == 2) {
      $userNameButton->set_active($true);
    }
  }
  if ($config->{"vowel"}) {
    $vowelButton->set_active($true);
  }
  my ($layout, $layout_arg) = @{$config->{"layout"}};
  if ($layout eq "columns") {
    $colButton->set_active($true);
    $colSpin->set_sensitive($true);
    $colSpin->set_value($layout_arg);
  }
  elsif ($layout eq "horizontal") {
    $horizButton->set_active($true);
  }
  elsif ($layout eq "vertical") {
    $vertButton->set_active($true);
  }
  elsif ($layout eq "oneclock") {
    $oneClockButton->set_active($true);
  }
  else {
    # nothing...
  }
  if ($config->{"format"}) {
    my $tf = $config->{"format"};
    $timeFormatCombo->entry->set_text($tf);
    _calculate_time_format_buttons($true);
  }
  # Needed because $tzList is cleared
  $selectedL = 0;
  if ($config->{"position"}) {
    $savePosButton->set_active($true);
  }
}

sub _ui_save_config {
  my ($main, $config) = @_;
  my ($options, @timezones) = "";
  my $m = $tzList->get_model();
  for (my $i = $m->get_iter_first(); defined $i; $i = $m->iter_next($i)) {
    push @timezones, [ $m->get($i, COLUMN_REAL_TIMEZONE),
		       $m->get($i, COLUMN_USERNAME) ];
  }
  return undef if ! @timezones;
  $config->{"zones"}      = \@timezones;
  $config->{"name"}
  = ($userNameButton->get_active() ?
     2 : ($nameButton ->get_active() ?
	  1 :
	  0));
  $config->{"vowel"}      = ($vowelButton->get_active() ? 1 : 0);
  if ($colButton->get_active()) {
    $config->{"layout"} = [ "columns", $colSpin->get_value_as_int() ];
  }
  elsif ($horizButton->get_active()) {
    $config->{"layout"} = [ "horizontal" ];
  }
  elsif ($vertButton->get_active()) {
    $config->{"layout"} = [ "vertical" ];
  }
  elsif ($oneClockButton->get_active()) {
    $config->{"layout"} = [ "oneclock" ];
  }
  $config->{"format"}     = $timeFormatCombo->entry->get_text();

  Intclock::Config::save($config);
  return 1;
}

1;
